# Copyright 2021 Facundo Batista
# Licensed under the GPL v3 License
# For further info, check https://github.com/facundobatista/pyempaq

"""Init file to allow execution of PyEmpaq as a module."""

import sys

from pyempaq import main

sys.exit(main.main())
